@@ -1,9 +1,10 @@
 import pandas as pd
 import numpy as np
 import matplotlib.pyplot as plt
-
+import matplotlib as mpl
+from matplotlib.dates import DateFormatter
 # ggplot stuff
-from components import colors, shapes, aes
+from components import colors, shapes, linestyles, aes
 from components.legend import draw_legend
 from geoms import *
 from scales import *
@@ -42,15 +43,14 @@ class ggplot(object):
     print p + geom_point()
     """
 
-    # TODO: make color discrete and continuous
-    CONTINUOUS = ['x', 'y', 'size']
-    DISCRETE = ['color', 'shape', 'marker', 'alpha']
+    CONTINUOUS = ['x', 'y', 'size', 'alpha']
+    DISCRETE = ['color', 'shape', 'marker', 'alpha', 'linestyle']
 
     def __init__(self, aesthetics, data):
         # ggplot should just 'figure out' which is which
         if not isinstance(data, pd.DataFrame):
             aesthetics, data = data, aesthetics
-            
+
         self.aesthetics = aesthetics
         self.data = data
 
@@ -71,7 +71,7 @@ class ggplot(object):
                             pass
                         else:
                             item = "self.data.get('%s')" % item
-                        lambda_column += item 
+                        lambda_column += item
                     self.data[name] = eval(lambda_column)
         # defaults
         self.geoms= []
@@ -85,6 +85,13 @@ class ggplot(object):
         self.title = None
         self.xlab = None
         self.ylab = None
+        # format for x/y major ticks
+        self.xtick_formatter = None
+        self.xbreaks = None
+        self.xtick_labels = None
+        self.xmajor_locator = None
+        self.xminor_locator = None
+        self.ytick_formatter = None
         self.xlimits = None
         self.ylimits = None
         self.legend = {}
@@ -94,12 +101,12 @@ class ggplot(object):
         self.colormap = plt.cm.Blues
 
     def __repr__(self):
-        # TODO: Handle facet_wrap better so that we only have
-        # as many plots as we do discrete values. Currently it
-        # creates a grid of plots but doesn't use all of them
+        # TODO: Handle facet_grid better. Currently facet_grid doesn't
+        # fuse the plots together, they're individual subplots.
         if self.facet_type=="grid":
-            fig, axs = plt.subplots(self.n_wide, self.n_high, 
+            fig, axs = plt.subplots(self.n_wide, self.n_high,
                     sharex=True, sharey=True)
+            plt.subplots_adjust(wspace=.05, hspace=.05)
         elif self.facet_type=="wrap":
             subplots_available = self.n_wide * self.n_high
             extra_subplots = subplots_available - self.n_dim_x
@@ -125,33 +132,66 @@ class ggplot(object):
         # dimensions of the plot remain the same
         if self.facets:
             cntr = 0
-            for facet, frame in self.data.groupby(self.facets):
-                for layer in self._get_layers(frame):
-                    for geom in self.geoms:
-                        if self.facet_type=="wrap":
-                            if cntr+1 > len(plots):
-                                continue
-                            pos = plots[cntr]
-                            if pos is None:
-                                continue
-                            y_i, x_i = pos
-                            pos = x_i + y_i * self.n_high + 1
-                            plt.subplot(self.n_wide, self.n_high, pos) 
-                        else:
-                            plt.subplot(self.n_wide, self.n_high, cntr)
-                            # TODO: this needs some work
-                            if (cntr % self.n_high)!=0:
-                                plt.tick_params(axis='y', which='both', 
-                                        bottom='off', top='off',
-                                        labelbottom='off')
+            if len(self.facets)==2:
+                for facets, frame in self.data.groupby(self.facets):
+                    pos = self.facet_pairs.index(facets) + 1
+                    plt.subplot(self.n_wide, self.n_high, pos)
+                    for layer in self._get_layers(frame):
+                        for geom in self.geoms:
+                            callbacks = geom.plot_layer(layer)
+                # This needs to enumerate all possibilities
+                for pos, facets in enumerate(self.facet_pairs):
+                    pos += 1
+                    if pos <= self.n_high:
+                        plt.subplot(self.n_wide, self.n_high, pos)
+                        plt.table(cellText=[[facets[1]]], loc='top',
+                                cellLoc='center', cellColours=[['lightgrey']])
+                    if pos % self.n_high==0:
+                        plt.subplot(self.n_wide, self.n_high, pos)
+                        x = max(plt.xticks()[0])
+                        y = max(plt.yticks()[0])
+                        ax = axs[pos % self.n_wide][pos % self.n_high]
+                        plt.table(cellText=[[facets[0]]], loc='right', 
+                            cellLoc='center', cellColours=[['lightgrey']])
+                    plt.subplot(self.n_wide, self.n_high, pos)
 
-                        callbacks = geom.plot_layer(layer)
-                        if callbacks:
-                            for callback in callbacks:
-                                fn = getattr(axs[cntr], callback['function'])
-                                fn(*callback['args'])
-                plt.title(facet)
-                cntr += 1
+                    if pos % self.n_high!=1:
+                        ticks = plt.yticks()
+                        plt.yticks(ticks[0], [])
+
+                    if pos <= (len(self.facet_pairs) - self.n_high):
+                        ticks = plt.xticks()
+                        plt.xticks(ticks[0], [])
+
+                    
+            else:
+                for facet, frame in self.data.groupby(self.facets):
+                    for layer in self._get_layers(frame):
+                        for geom in self.geoms:
+                            if self.facet_type=="wrap":
+                                if cntr+1 > len(plots):
+                                    continue
+                                pos = plots[cntr]
+                                if pos is None:
+                                    continue
+                                y_i, x_i = pos
+                                pos = x_i + y_i * self.n_high + 1
+                                plt.subplot(self.n_wide, self.n_high, pos)
+                            else:
+                                plt.subplot(self.n_wide, self.n_high, cntr)
+                                # TODO: this needs some work
+                                if (cntr % self.n_high)==-1:
+                                    plt.tick_params(axis='y', which='both',
+                                            bottom='off', top='off',
+                                            labelbottom='off')
+                            callbacks = geom.plot_layer(layer)
+                            if callbacks:
+                                for callback in callbacks:
+                                    fn = getattr(axs[cntr], callback['function'])
+                                    fn(*callback['args'])
+                    #TODO: selective titles
+                    plt.title(facet)
+                    cntr += 1
         else:
             for layer in self._get_layers(self.data):
                 for geom in self.geoms:
@@ -161,15 +201,32 @@ class ggplot(object):
                         for callback in callbacks:
                             fn = getattr(axs, callback['function'])
                             fn(*callback['args'])
-        
-        # Handling the details of the chart here; might be a better
+
+        # Handling the details of the chart here; probably be a better
         # way to do this...
         if self.title:
             plt.title(self.title)
         if self.xlab:
-            plt.xlabel(self.xlab)
+            if self.facet_type=="grid":
+                fig.text(0.5, 0.025, self.xlab)
+            else:
+                plt.xlabel(self.xlab)
         if self.ylab:
-            plt.ylabel(self.ylab)
+            if self.facet_type=="grid":
+                fig.text(0.025, 0.5, self.ylab, rotation='vertical')
+            else:
+                plt.ylabel(self.ylab)
+        if self.xmajor_locator:
+            plt.gca().xaxis.set_major_locator(self.xmajor_locator)
+        if self.xtick_formatter:
+            plt.gca().xaxis.set_major_formatter(self.xtick_formatter)
+            fig.autofmt_xdate()
+        if self.xbreaks: # xbreaks is a list manually provided
+            plt.gca().xaxis.set_ticks(self.xbreaks)
+        if self.xtick_labels:
+            plt.gca().xaxis.set_ticklabels(self.xtick_labels)
+        if self.ytick_formatter:
+            plt.gca().yaxis.set_major_formatter(self.ytick_formatter)
         if self.xlimits:
             plt.xlim(self.xlimits)
         if self.ylimits:
@@ -181,8 +238,7 @@ class ggplot(object):
         if self.legend:
             if self.facets:
                 if 1==2:
-                    ax_to_use = [ax for ax in axs if not isinstance(ax, np.ndarray)]
-                    ax = axs[-1]
+                    ax = axs[0][self.n_wide]
                     box = ax.get_position()
                     ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])
                     cntr = 0
@@ -198,8 +254,8 @@ class ggplot(object):
                         axs.add_artist(draw_legend(axs, legend, name, cntr))
                         cntr += 1
 
-        # TODO: We can probably get pretty sugary with this
-        return "<ggplot: (%d)>" % self.__hash__() 
+        # TODO: We can probably get more sugary with this
+        return "<ggplot: (%d)>" % self.__hash__()
 
     def _get_layers(self, data=None):
         # This is handy because... (something to do w/ facets?)
@@ -252,6 +308,15 @@ class ggplot(object):
             mapping['marker'] = mapping['shape'].replace(shape_mapping)
             del mapping['shape']
 
+        rev_linetype_mapping = {}
+        if 'linestyle' in mapping:
+            mapping['linestyle'] = mapping['linestyle'].apply(str)
+            possible_styles = np.unique(mapping['linestyle'])
+            linestyle = linestyles.line_gen()
+            line_mapping = {value: linestyle.next() for value in possible_styles}
+            rev_linetype_mapping = {v: k for k, v in line_mapping.iteritems()}
+            mapping['linestyle'] = mapping['linestyle'].replace(line_mapping)
+
         keys = [ae for ae in self.DISCRETE if ae in mapping]
         if "cmap" in mapping:
             keys.remove("color")
@@ -270,7 +335,7 @@ class ggplot(object):
                 legend["color"] = dict(zip(key_colors, quantiles))
             layers.append(frame)
         else:
-            legend = {"color": {}, "marker": {}}
+            legend = {"color": {}, "marker": {}, "linestyle": {}}
             for name, frame in mapping.groupby(keys):
                 frame = frame.to_dict('list')
                 for ae in self.DISCRETE:
@@ -287,10 +352,19 @@ class ggplot(object):
                             legend[ae][frame[ae]] = rev_shape_mapping.get(aes_name, aes_name)
                             # raise Exception("Cannot have numeric shape!")
                             label = rev_shape_mapping.get(aes_name, aes_name)
+                        elif ae=="linestyle":
+                            legend[ae][frame[ae]] = rev_linetype_mapping.get(aes_name, aes_name)
+                            label = rev_linetype_mapping.get(aes_name, aes_name)
+                        elif ae=="alpha":
+                            label = ""
                         frame['label'] = label
                 if "cmap" in frame:
                     frame["cmap"] = frame["cmap"][0]
                 layers.append(frame)
+        if "size" in mapping:
+            labels = np.percentile(mapping['size'], [5, 25, 50, 75, 95])
+            quantiles = np.percentile(self.data[self.aesthetics['size']], [5, 25, 50, 75, 95])
+            legend["size"] = dict(zip(labels, quantiles))
         # adding legends back to the plot
         self.legend = legend
         return layers
